/*
   Unix SMB/CIFS implementation.
   Samba Web Administration Tool
   Version 3.0.0
   Copyright (C) Andrew Tridgell 1997-2002
   Copyright (C) John H Terpstra 2002

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
/*
 * 1997 - 2013    SWAT
 *       R.I.P.                                                    :,
 *  Finally swatted, you will bug us no more.                     @,@
 *                                                              +#`@
 *                                                             @`:@
 *              ,'                                           :@ #'
 *             +   @                                        @: @.
 *            @    +                                      ,@ .@
 *           ;`     ;                                    @; #+
 *   @@`     @      @                                  ;@  @`
 *  #  :@   '       #                                 @, ##
 * @     +.  #       `                              ## .@
 *         #  #      '                            ;@  @;
 *,         '  #     +                          ,@` @@
 *'           ` +    :                        .@, #@
 *#       `@@#.  :                          ;@: '@
 *+.       #` ,#@@@@@@@@@@@@@             #@` @@
 * .@@.    @ `#@+,`    `##@+ +         .@@ `@#
 *     +@. @ ;@;. `:;#   '`:+@      `@@, +@:
 *        :@,#,  +`:;;;#.@@.:'   :@@: '@#
 *         + ,: '+@@@@@@@@@+ ;@@+  '@@
 *         .: :@@;.  @: ;;+@@; ;@@'
 *          # ,#@',  @# .`;#@ @
 *          #  .@     ''    ; @
 *          @  ,:@@@+;++..+,. @
 *          @  @+;+@+@, .  ;` @'
 *          @  :         . @,'@ '
 *         .@@+#+''''++#@@@+;.,
 *         .@ @'.    :  :++@@# .
 *        + @:#@;'+@@@@@#;
 *            ` :@;   .@ '#
 *                 ,:    ,#
 *               ,        +.
 *           ,   :
 *                `
 *                +
 */
