/* 
   Unix SMB/CIFS implementation.

   Copyright (C) Stefan Metzmacher 2007

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef _NTVFS_COMMON_H_
#define _NTVFS_COMMON_H_

#include "ntvfs/ntvfs.h"

struct notify_event;
struct notify_entry;

#include "ntvfs/common/brlock.h"
#include "ntvfs/common/opendb.h"
#include "ntvfs/common/proto.h"

#endif /* _NTVFS_COMMON_H_ */
