/*-
 * Copyright (c) 2005 Doug Rabson
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	$FreeBSD: src/lib/libgssapi/context.h,v 1.1 2005/12/29 14:40:20 dfr Exp $
 *	$Id$
 */

#include <gssapi_mech.h>

struct _gss_context {
        gss_buffer_desc          gc_input;
        char                    *gc_free_this;
        size_t                   gc_target_len;
        size_t                   gc_oid_offset;
        gssapi_mech_interface    gc_mech;
        gss_ctx_id_t             gc_ctx;
        uint8_t                  gc_initial;
};

#define EXPORT_CONTEXT_VERSION_MASK             0x03
#define EXPORT_CONTEXT_FLAGS_MASK               0xfc
#define EXPORT_CONTEXT_FLAG_ACCUMULATING        0x04
#define EXPORT_CONTEXT_FLAG_MECH_CTX            0x08

void
_gss_mg_error(gssapi_mech_interface, OM_uint32);

OM_uint32
_gss_mg_get_error(const gss_OID, OM_uint32, gss_buffer_t);
